from rich.text import TextType
from textual.app import App, ComposeResult
from textual.widgets import (
    Footer,
    Header,
    Static,
    Button,
    ContentSwitcher,
)
from textual.containers import (
    VerticalScroll,
    Horizontal,
    Vertical,
    Container as Group,
)
from textual.reactive import reactive
from docker import DockerClient
from docker.models.containers import Container

from .utils import get_cpu_usage, get_mem_usage


class AppGUI(App):
    CSS_PATH = "style.css"
    BINDINGS = [
        ("h", "home", "Home"),
        ("d", "toggle_dark", "Toggle dark mode"),
        ("q", "quit", "Quit"),
    ]
    TITLE = "DOCKERY"

    def __init__(self, docker: DockerClient, **kargs):
        self.docker = docker
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield Footer()
        with ContentSwitcher(initial="container-list"):
            yield VerticalScroll(id="container-list")
            with VerticalScroll(id="container-logs"):
                yield Static("", id="logs")

    def on_mount(self) -> None:
        cl = self.query_one("#container-list")
        for c in self.docker.containers.list(all=True):
            cw = ContainerWidget(c, self.docker)  # type: ignore
            cl.mount(cw)

    def action_home(self) -> None:
        self.query_one(ContentSwitcher).current = "container-list"


class ContainerWidget(Static):
    def __init__(self, container: Container, client: DockerClient, **kargs):
        self.container = container
        self.client = client
        self.container_id = container.id
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Group(
            Static(self.container.name or ""),
            Static(self.container.short_id),
            classes="container-info",
        )
        yield Vertical(
            ReactiveString(id="status"),
            ReactiveString(id="cpu"),
            ReactiveString(id="mem"),
            classes="status-text",
        )
        yield Group(StatusButtons(self.container))

    def on_mount(self):
        self.set_interval(1, self.update_data)
        # self.set_interval(4, self.update_usage) # TODO: is very slow

    def update_data(self) -> None:
        c: Container = self.client.containers.get(self.container_id)  # type: ignore
        self.container = c
        self.query_one("#status", ReactiveString).value = c.status.capitalize()

    async def update_usage(self) -> None:
        c = self.container
        cpu = self.query_one("#cpu", ReactiveString)
        mem = self.query_one("#mem", ReactiveString)
        if c.status == "running":
            stats = c.stats(decode=None, stream=False)  # slow af
            cpu.value = f"CPU: {get_cpu_usage(stats):.1f}"
            mem.value = f"MEM: {get_mem_usage(stats):.1f}"
        else:
            cpu.value = "CPU: -"
            mem.value = "MEM: -"


class ReactiveString(Static):
    value = reactive("")

    def render(self) -> str:
        return self.value


class StatusButtons(Static):
    def __init__(self, container: Container, **kargs):
        self.container = container
        self.button = StartStopButton(container)
        super().__init__(**kargs)

    def compose(self) -> ComposeResult:
        yield Horizontal(
            self.button,
            RestartButton(self.container),
            LogsButton(self.container),
            # Button("Details"),
            classes="buttons-container",
        )


class StartStopButton(Button):
    def __init__(self, container: Container, **kargs):
        self.running = container.status == "running"
        self.container = container
        super().__init__(**kargs)

    def on_click(self) -> None:
        self.running = not self.running
        if self.running:
            self.container.start()
        else:
            self.container.stop()
        self.on_mount()

    def on_mount(self) -> None:
        self.variant = "error" if self.running else "success"
        self.text = "Stop" if self.running else "Start"

    def render(self) -> TextType:
        return self.text


class RestartButton(Button):
    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(**kargs)

    def on_click(self) -> None:
        self.container.restart()

    def on_mount(self) -> None:
        self.variant = "warning"

    def render(self) -> TextType:
        return "Restart"


class LogsButton(Button):
    def __init__(self, container: Container, **kargs):
        self.container = container
        super().__init__(**kargs)

    def on_click(self) -> None:
        logs: bytes = self.container.logs(tail=20)
        cl = self.app.query_one("#logs")
        lw = Static(logs.decode())
        cl.mount(lw)
        self.app.query_one(ContentSwitcher).current = "container-logs"

    def on_mount(self) -> None:
        self.variant = "primary"

    def render(self) -> TextType:
        return "Logs"



import docker
from docker.models.containers import Container
from docker.models.volumes import Volume
from docker.models.images import Image
import click
from rich import print

from .gui import AppGUI

default_options = [
    click.option(
        "--server",
        default=None,
        help="Docker server url, for example: unix:///var/run/docker.sock",
    ),
    click.option("--ssh", is_flag=True),
    click.option("--api-version", default="1.35", help="Docker API version"),
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def get_client(server, ssh, api_version, **kargs):
    if server is None:
        client = docker.from_env()
    else:
        client = docker.DockerClient(
            base_url=server, version=api_version, use_ssh_client=ssh
        )
    return client


def run_gui(**kargs):
    client = get_client(**kargs)
    gui = AppGUI(client)
    gui.run()


@click.group(invoke_without_command=True)
@add_options(default_options)
@click.pass_context
def main(ctx, **kargs):
    if ctx.invoked_subcommand is None:
        run_gui(**kargs)


@click.command
@add_options(default_options)
def df(**kargs):
    client = get_client(**kargs)
    print(client.df())


@click.command
@add_options(default_options)
def volumes(**kargs):
    client = get_client(**kargs)
    vlms: list[Volume] = client.volumes.list()  # type: ignore
    for v in vlms:
        print(v.attrs)


@click.command
@add_options(default_options)
@click.option("--all", "-a", is_flag=True)
def ps(**kargs):
    client = get_client(**kargs)
    containers: list[Container] = client.containers.list(
        all=kargs["all"],
    )  # type: ignore
    for c in containers:
        print(c.attrs)


@click.command
@add_options(default_options)
def stats(**kargs):
    client = get_client(**kargs)
    containers: list[Container] = client.containers.list(all=True)  # type: ignore
    for c in containers:
        print(f"[b]{c.name}:")
        stats = c.stats(decode=None, stream=False)
        print(stats)


@click.command
@add_options(default_options)
def images(**kargs):
    client = get_client(**kargs)
    imgs: list[Image] = client.images.list()  # type: ignore
    for i in imgs:
        print(i.attrs)


main.add_command(df)
main.add_command(volumes)
main.add_command(ps)
main.add_command(stats)
main.add_command(images)

if __name__ == "__main__":
    main()



def get_cpu_usage(stats: dict) -> float:
    cpu_stats = stats["cpu_stats"]
    precpu_stats = stats["precpu_stats"]
    delta = (
        cpu_stats["cpu_usage"]["total_usage"] - precpu_stats["cpu_usage"]["total_usage"]
    )
    system_delta = cpu_stats.get("system_cpu_usage", 0) - precpu_stats.get(
        "system_cpu_usage", 0
    )
    len_cpu = cpu_stats.get("online_cpus", 1)

    if system_delta == 0:
        return 0.0

    percentage = (delta / system_delta) * len_cpu * 100
    return percentage


def get_mem_usage(stats: dict) -> float:
    mem_stats = stats["memory_stats"]
    if len(mem_stats) == 0:
        return 0.0
    mem_used = (
        mem_stats["usage"]
        - mem_stats["stats"].get("cache", 0)
        + mem_stats["stats"]["active_file"]
    )
    limit = stats["memory_stats"]["limit"]
    percentage = mem_used / limit * 100
    return percentage
